#!/usr/bin/python

### Author: Marcus Falls marcus.falls@bsc.es
### Regrids input data in a regular grid to an orca one.

# README:
# This can be run on the local machine or nord3.
# It's set up to only work on the files that fit the /esarchive conventions; i.e appropriate paths, one monthly file
#       per variable.
# It's set up to work on directories.
# If the input contains no time data, it will take the 15th of the month of each input file as the time.
# Ensure that you have access to the orca1 data: domain_cfg.nc and coordinates_xios.nc.
#
# Instructions for use:
# module load Iris
# python to_ORCA.py var /path/to/monthly/
#       The inputs are in /path/to/monthly/var/ and the outputs will be found in /path/to/monthly/var-orca1/
# e.g python to_ORCA.py chlos /esarchive/obs/esa/occci-v5.0/monthly_mean/
#       Inputs: /esarchive/obs/esa/occci-v5.0/monthly_mean/chlos/ /esarchive/obs/esa/occci-v5.0/monthly_mean/chlos-orca1/


import numpy as np
from netCDF4 import Dataset
from scipy import interpolate
import os
import sys
import iris
from datetime import date,time
import math

def expand(gr, da):
    grnew = np.zeros((len(gr)+2))
    grnew[1:-1] = gr[:]
    grnew[0] = 2*grnew[1] - grnew[2]
    grnew[-1] = 2*grnew[-2] - grnew[-3]
    if da.ndim == 2:
        danew = np.zeros((np.shape(da)[0], np.shape(da)[1] + 2))
        danew[:,1:-1] = da
        danew[:,0] = da[:,-1]
        danew[:,-1] = da[:,0]
    elif da.ndim == 3:
        danew = np.zeros((np.shape(da)[0], np.shape(da)[1],  np.shape(da)[2] + 2))
        danew[:,:,1:-1] = da
        danew[:,:,0] = da[:,:,-1]
        danew[:,:,-1] = da[:,:,0]
    else:
        raise()
    return grnew, danew

def get_interp(varname, inputfile, orcafile, boundsfile, outputfile):

    # extract data from the variable file
    indata = Dataset(inputfile)
    inlatname = list(set(indata.variables.keys()).intersection(['lat','latitude']))[0]
    inlonname = list(set(indata.variables.keys()).intersection(['lon','longitude']))[0]
    inlat = np.array(indata[inlatname])
    inlon = np.array(indata[inlonname])

    # if there is a time dimension and variable
    if 'time' in indata.dimensions.keys() or 't' in indata.dimensions.keys():
        var = np.array(indata[varname])[0]
    # if there is not - approximate the time as the 15th of the month on the file
    else:
        var = np.array(indata[varname])

    var[var>100000000] = -1.
    var[var==-999] = -1.
    var[np.isnan(var)] = -1

    inlon, var = expand(inlon,var) # allow for interpolation at the map edges
    if inlat[0]>inlat[-1]: # arrange the lats and var as ascending, if need be
        inlat=inlat[::-1]
        if var.ndim == 2:
            var = var[::-1]
        elif var.ndim == 3:
            var = var[:,::-1,:]

    # extract data from the orca file
    orcadata = Dataset(orcafile)
    orcalat = np.array(orcadata['nav_lat'])
    orcalon = np.array(orcadata['nav_lon'])
    (y,x) = np.shape(orcalat)
    if np.max(inlon>300):
        orcalon[np.where(orcalon < 0)] = orcalon[np.where(orcalon < 0)] + 360 # changing lon range to 0-360 for easier interpolation
    orcagrid = np.swapaxes(np.swapaxes(np.array([orcalat, orcalon]), 0, 2), 0, 1)

    xi = np.zeros((np.size(orcalat),2))
    (Y,X) = np.meshgrid(inlat,inlon)
    points = np.zeros((np.size(X),2))

    xi[:,-2] = np.ravel(orcalat)
    xi[:,-1] = np.ravel(orcalon)
    points[:,-2] = np.ravel(Y)
    points[:,-1] = np.ravel(X)

    var1d = np.ravel(var)

    # interpolation 1 - filling the gaps in the original data
    if var.ndim == 2:
        varfilled = interpolate.griddata(points[var1d > 0], var1d[var1d > 0], points, method='cubic')
        varfilled = interpolate.griddata(points[varfilled>0], varfilled[varfilled>0], points, method='nearest')
        varfilled = np.reshape(varfilled, (len(inlat),len(inlon)))
        orcavar = interpolate.interpn((inlat, inlon), varfilled, orcagrid, method="splinef2d")
    if var.ndim == 3:
        indepth = np.array(indata['depth']).astype(float)
        new2var = np.zeros((len(indepth),y,x))

        origyx = len(inlat)*len(inlon)
        def interp_layer(k):
            s = origyx*k
            e = origyx*(k+1)
            if len(var1d[s:e][var1d[s:e] > 0])>3:
                varlayerfilled = interpolate.griddata(points[var1d[s:e] > 0], var1d[s:e][var1d[s:e] > 0], points, method='cubic')
                varlayerfilled = interpolate.griddata(points[varlayerfilled>0], varlayerfilled[varlayerfilled>0], points, method='nearest')
                varlayerfilled = np.reshape(varlayerfilled, (len(inlat),len(inlon)))
                varlayerfilled = interpolate.interpn((inlat, inlon), varlayerfilled, orcagrid, method="splinef2d")
            else: # for the layers that have little or no data
                varlayerfilled = np.zeros(np.shape(orcalat)).astype('float32')
                varlayerfilled[:] = np.nan
            new2var[k] = varlayerfilled

        vinterp_layer = np.vectorize(interp_layer)
        vinterp_layer(range(len(indepth)))

        orcadepth = np.cumsum(np.array(orcadata['e3w_1d']))
        new3var = np.zeros((len(orcadepth),y,x))

        def vert_interp(k): #vertical interpolation
            i = int(math.floor(k/y))
            j = k%y
            kvar = new2var[:,j,i][~np.isnan(new2var[:,j,i])]
            #print(kvar)
            f = interpolate.interp1d(indepth[:len(kvar)], kvar, kind='cubic')
            new_column = f(orcadepth[orcadepth < np.max(indepth[:len(kvar)])])
            #print(new_column)
            new3var[:len(new_column),j,i] = new_column

        v_vert_interp = np.vectorize(vert_interp)
        v_vert_interp(range(np.size(orcalat)))

        orcavar = new3var

    # removing points surrounded by nans in the original data, and also land
    if var.ndim == 2:
        for j in range(np.shape(orcavar)[0]):
            for i in range(np.shape(orcavar)[1]):
                boundj = np.where(inlat==np.max(inlat[inlat < orcalat[j, i]]))[0][0]
                boundi = np.where(inlon==np.max(inlon[inlon < orcalon[j, i]]))[0][0]
                if var[boundj,boundi]==-1 and var[boundj+1,boundi]==-1 and var[boundj,boundi+1]==-1 and var[boundj+1,boundi+1]==-1:
                    orcavar[j,i]=np.nan
    elif var.ndim == 3:
        for k in range(np.shape(orcavar)[0]):
            if orcadepth[k] < indepth[-1]:
                boundk = np.where(indepth == np.max(indepth[indepth < orcadepth[k]]))[0][0]
                for j in range(np.shape(orcavar)[1]):
                    for i in range(np.shape(orcavar)[2]):
                        boundj = np.where(inlat==np.max(inlat[inlat < orcalat[j, i]]))[0][0]
                        boundi = np.where(inlon==np.max(inlon[inlon < orcalon[j, i]]))[0][0]
                        if var[boundk,boundj,boundi]==-1 and var[boundk,boundj+1,boundi]==-1 and var[boundk,boundj,boundi+1]==-1 and var[boundk,boundj+1,boundi+1]==-1\
                                and var[boundk+1,boundj,boundi]==-1 and var[boundk+1,boundj+1,boundi]==-1 and var[boundk+1,boundj,boundi+1]==-1 and var[boundk+1,boundj+1,boundi+1]==-1:

                            orcavar[k,j,i]=np.nan

    # removing land
    mask = np.array(orcadata['bottom_level'])[0]
    if var.ndim==2:
        orcavar[mask == 0] = np.nan
    elif var.ndim==3:
        for j in range(np.shape(orcavar)[0]):
            for i in range(np.shape(orcavar)[1]):
                orcavar[mask[j,i]:,j,i] = np.nan

    # removing neg values generated by the interpolation
    orcavar[orcavar <= 0] = np.nan

    # bounds
    boundsdata = Dataset(boundsfile)
    latbnds = np.array(boundsdata['bounds_nav_lat_grid_T'])
    lonbnds = np.array(boundsdata['bounds_nav_lon_grid_T'])

    if var.ndim == 3:
        depthbnds = np.zeros((len(orcadepth),2)).astype(float)
        between_depths = (orcadepth[1:] + orcadepth[:-1])/2
        depthbnds[1:,0] = between_depths
        depthbnds[:-1,1] = between_depths
        depthbnds[-1,-1] = 2*orcadepth[-1] - between_depths[-1]

    # creation of the iris cube

    if var.ndim == 3:
        lev = iris.coords.DimCoord(
            orcadepth,
            var_name='lev',
            long_name='ocean depth coordinate',
            units = 'm',
            bounds = depthbnds,
            standard_name='depth',
            attributes={'positive': 'down'}
        )
        lev.axis = 'Z'


    j_coord = iris.coords.DimCoord(
        range(y),
        var_name='j',
        long_name='cell index along second dimension',
    )

    i_coord = iris.coords.DimCoord(
        range(x),
        var_name='i',
        long_name='cell index along first dimension',
    )
    lat_coord = iris.coords.AuxCoord(
        orcalat,
        bounds=latbnds,
        var_name='lat',
        standard_name='latitude',
        units='degrees_north',
    )
    lon_coord = iris.coords.AuxCoord(
        orcalon,
        bounds=lonbnds,
        var_name='lon',
        standard_name='longitude',
        units='degrees_east',
    )

    try:
        intime = np.array(indata['time'])
        time_coord = iris.coords.DimCoord(
            intime,
            var_name='time',
            standard_name=indata['time'].standard_name,
            units=indata['time'].units,
        )
    # if there is no time data, one is created using the 15th day of the month
    except:
        delta = date(int(inputfile[-9:-5]),int(inputfile[-5:-3]), 15)- date(1970, 1, 1)
        intime = np.array([delta.days])
        time_coord = iris.coords.DimCoord(
            intime,
            var_name='time',
            standard_name='time',
            units='days since 1970-01-01 00:00:00',
        )

    # Create tos variable cube
    if var.ndim == 2:
        datanew = iris.cube.Cube(
            np.array([orcavar]),
            var_name=varname,
            long_name=indata[varname].long_name,
            units=indata[varname].units,
            dim_coords_and_dims=[(time_coord, 0), (j_coord, 1), (i_coord, 2)],
            aux_coords_and_dims=[(lat_coord, (1, 2)), (lon_coord, (1, 2))],
        )

    if var.ndim == 3:
        datanew = iris.cube.Cube(
            np.array([orcavar]),
            var_name=varname,
            long_name=indata[varname].long_name,
            units=indata[varname].units,
            dim_coords_and_dims=[(time_coord, 0), (lev, 1), (j_coord, 2), (i_coord, 3)],
            aux_coords_and_dims=[(lat_coord, (2, 3)), (lon_coord, (2, 3))],
        )

    try:
        datanew.standard_name=indata[varname].standard_name
    except:
        datanew.standard_name = None

    iris.save(datanew,outputfile)

    return datanew

variable = sys.argv[1]
path = sys.argv[2]

inpath = path + variable + '/'
outpath = path + variable + '-orca1/'

orca1file = '/esarchive/scratch/mfalls/domain_cfg.nc'
bfile = '/esarchive/scratch/mfalls/test/coordinates_xios.nc'

files = os.listdir(inpath)
files.sort()

for f in files:
    print('Processing '+f)
    if os.path.exists(outpath+f):
        os.remove(outpath+f)
    get_interp(variable, inpath+f, orca1file, bfile, outpath+f)

# print('test')
# get_interp('chl', '/esarchive/obs/esa/pc_oc-cci-v5/monthly/chlos/chl_201502.nc', '/esarchive/scratch/mfalls/domain_cfg.nc','/esarchive/scratch/mfalls/test/coordinates_xios.nc','/esarchive/obs/esa/pc_oc-cci-v5/monthly/chlos-orca1/chlos_201502.nc')

